#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/pci.h>
#include <linux/etherdevice.h>
#include <linux/mii.h>
#include <linux/delay.h>
#include <linux/ethtool.h>

#include "et131x.h"
#include "et131x_phy.h"
#include "et131x_dma.h"

#define DRV_NAME	"et131x"
#define DRV_VERSION	"0.1"
#define DRV_DESCRIPTION	"Agere Systems Inc. ET131x Gigabit Ethernet driver"
#define DRV_COPYRIGHT	"(C) 2006 Guus Sliepen <guus@sliepen.eu.org>"
#define DRV_LICENSE     "GPL"

static const char testpacket[] = {
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0x0, 0x1, 0x2, 0x3, 0x4, 0x5,
	0x08, 0x00,
	0x45, 0x00,
        0x00, 0x54,  0x00, 0x00,  0x40, 0x00,  0x40, 0x01,  0x24, 0xa5,  0x01, 0x02,  0x03, 0x04,  0xff, 0xff,
        0xff, 0xff,  0x08, 0x00,  0x95, 0xea,  0xcb, 0x11,  0x00, 0x01,  0xc5, 0x2b,  0x02, 0x44,  0xda, 0x8f,
        0x0a, 0x00,  0x08, 0x09,  0x0a, 0x0b,  0x0c, 0x0d,  0x0e, 0x0f,  0x10, 0x11,  0x12, 0x13,  0x14, 0x15,
        0x16, 0x17,  0x18, 0x19,  0x1a, 0x1b,  0x1c, 0x1d,  0x1e, 0x1f,  0x20, 0x21,  0x22, 0x23,  0x24, 0x25,
        0x26, 0x27,  0x28, 0x29,  0x2a, 0x2b,  0x2c, 0x2d,  0x2e, 0x2f,  0x30, 0x31,  0x32, 0x33,  0x34, 0x35,
};

void et131x_showregs(struct et131x *et131x) {
	printk(KERN_DEBUG "et13x: global register pm_ctrl = %08x\n", et131x->regs->global.pm_ctrl.value);
	printk(KERN_DEBUG "et13x: global register pm_state = %08x\n", et131x->regs->global.pm_state.value);
	printk(KERN_DEBUG "et13x: global register int_status = %08x\n", et131x->regs->global.int_status.value);
	printk(KERN_DEBUG "et13x: global register int_mask = %08x\n", et131x->regs->global.int_mask.value);
	printk(KERN_DEBUG "et13x: global register msi_config = %08x\n", et131x->regs->global.msi_config.value);
	printk(KERN_DEBUG "et13x: global register loopback = %08x\n", et131x->regs->global.loopback.value);

	printk(KERN_DEBUG "et13x: txdma register ctrl = %08x\n", et131x->regs->txdma.ctrl.value);
	printk(KERN_DEBUG "et13x: rxdma register ctrl = %08x\n", et131x->regs->rxdma.ctrl.value);
	printk(KERN_DEBUG "et13x: txmac register ctrl = %08x\n", et131x->regs->txmac.ctrl.value);
	printk(KERN_DEBUG "et13x: rxmac register ctrl = %08x\n", et131x->regs->rxmac.ctrl.value);

	printk(KERN_DEBUG "et13x: mac register ctrl1 = %08x\n", et131x->regs->mac.ctrl1.value);
	printk(KERN_DEBUG "et13x: mac register ctrl2 = %08x\n", et131x->regs->mac.ctrl2.value);
	printk(KERN_DEBUG "et13x: mac register ipg = %08x\n", et131x->regs->mac.ipg.value);
	printk(KERN_DEBUG "et13x: mac register max_fm_len = %08x\n", et131x->regs->mac.max_fm_len);

	printk(KERN_DEBUG "et13x: mmc register ctrl = %08x\n", et131x->regs->mmc.ctrl.value);
}

static irqreturn_t et131x_intr(int irq, void *dev_id, struct pt_regs *regs) {
	struct et131x *et131x = dev_id;
	union et131x_int int_mask;
	union et131x_int int_status;
	static int count = 0;

	if(!et131x)
		return IRQ_NONE;

	int_mask = et131x->regs->global.int_mask;
	et131x->regs->global.int_mask.value = ~0;
	wmb();
	{
		int status, speed, duplex, autoneg;
		et131x_phy_linkstatus(et131x, &status, &autoneg, &speed, &duplex, NULL, NULL, NULL);
		printk(KERN_DEBUG "et131x: linkstatus %d speed %d duplex %d autoneg %d\n", status, speed, duplex, autoneg);
	}

	int_status = et131x->regs->global.int_status;

	// printk(KERN_DEBUG "et131x: got interrupt %d for device %s, status %08x, count %d\n", irq, et131x->netdev->name, int_status.value, count++);
	// TODO: Schedule actual handling.

	et131x->regs->global.int_mask = int_mask;
	return int_status.value ? IRQ_HANDLED : IRQ_NONE;
}

static int __devinit et131x_probe(struct pci_dev *pcidev, const struct pci_device_id *pciid) {
	int err = 0;
	struct et131x *et131x;
	struct net_device *netdev;
	uint8_t eeprom_status;
	uint8_t pci_revid;
	uint8_t macaddr[6];
	int i;
	
	if(!pcidev)
		return -ENODEV;

	// Allocate etherdev early so we also have a struct et131x.
	if(!(netdev = alloc_etherdev(sizeof *et131x))) {
		printk(KERN_ERR "et131x: unable to allocate network device\n");
		err = -ENOMEM;
		goto exit;
	}

	et131x = netdev_priv(netdev);
	pci_set_drvdata(pcidev, et131x);

	et131x->pcidev = pcidev;
	et131x->netdev = netdev;

	printk(KERN_DEBUG "et131x: probing device %s\n", pci_name(pcidev));

	printk(KERN_DEBUG "et131x: enabling PCI device\n");
	if((err = pci_enable_device(pcidev))) {
		printk(KERN_ERR "et131x: unable to enable PCI device\n");
		goto exit_free_netdev;
	}

	if(!(pci_resource_flags(pcidev, 0) & IORESOURCE_MEM)) {
		printk(KERN_ERR "et131x: resource flags test failed\n");
		err = -ENODEV;
		goto exit_disable_device;
	}

	if((err = pci_request_regions(pcidev, DRV_NAME))) {
		printk(KERN_ERR "et131x: request regions test failed\n");
		goto exit_disable_device;
	}

	SET_MODULE_OWNER(netdev);
	SET_NETDEV_DEV(netdev, &pcidev->dev);

	pci_set_master(pcidev);

	if(!pci_find_capability(pcidev, PCI_CAP_ID_PM)) {
		printk(KERN_ERR "et131x: power management capabilities test failed\n");
		err = -ENODEV;
		goto exit_release_regions;
	}

	if(sizeof(dma_addr_t) > sizeof(u32) && !pci_set_dma_mask(pcidev, DMA_64BIT_MASK)) {
		if((err = pci_set_consistent_dma_mask(pcidev, DMA_64BIT_MASK))) {
			printk(KERN_ERR "et131x: DMA setup failed\n");
			goto exit_release_regions;
		}
		et131x->using_dac = 1;
	} else {
		if((err = pci_set_dma_mask(pcidev, DMA_32BIT_MASK))) {
			printk(KERN_ERR "et131x: DMA setup failed\n");
			goto exit_release_regions;
		}
		et131x->using_dac = 0;
	}

	// Read EEPROM
	if(pci_read_config_byte(pcidev, ET1310_PCI_EEPROM_STATUS, &eeprom_status)) {
		printk(KERN_ERR "et131x: unable to read EEPROM status\n");
		err = -ENODEV;
		goto exit_release_regions;
	}

	if(pci_read_config_byte(pcidev, PCI_REVISION_ID, &pci_revid)) {
		printk(KERN_ERR "et131x: unable to read revision id\n");
		err = -ENODEV;
		goto exit_release_regions;
	}

	printk(KERN_INFO "et131x: eeprom status %x, revision %d\n", eeprom_status, pci_revid);
	if(eeprom_status & 0x4C && pci_revid == 1)
		printk(KERN_INFO "et131x: card needs eeprom workaround\n");

	// TODO: write latency registers?

	for(i = 0; i < 6; i++) {
		if(pci_read_config_byte(pcidev, ET1310_PCI_MAC_ADDRESS + i, &macaddr[i])) {
			printk(KERN_ERR "et131x: unable to read MAC address from EEPROM\n");
			err = -ENODEV;
			goto exit_release_regions;
		}
	}
	printk(KERN_INFO "et131x: MAC address %02x:%02x:%02x:%02x:%02x:%02x\n",
			macaddr[0], macaddr[1], macaddr[2], macaddr[3], macaddr[4], macaddr[5]); 
	
	if(pci_resource_len(pcidev, 0) < sizeof *et131x->regs) {
		printk(KERN_ERR "et131x: PCI resource length smaller than known register set\n");
		err = -ENODEV;
		goto exit_release_regions;
	}

	memcpy(netdev->dev_addr, macaddr, ETH_ALEN);
	memcpy(netdev->perm_addr, macaddr, ETH_ALEN);

	et131x->regs = ioremap_nocache(pci_resource_start(pcidev, 0), sizeof *et131x->regs);
	if(!et131x->regs) {
		printk(KERN_ERR "et131x: unable to map device registers\n");
		err = -ENOMEM;
		goto exit_release_regions;
	}

	et131x_showregs(et131x);
	et131x_soft_reset(et131x);
	et131x_disable_interrupts(et131x);
	//et131x_wake(et131x);
	et131x_config_global(et131x);
	et131x_config_mac1(et131x);
	//et131x_config_mac2(et131x);
	{
		int status = et131x_xcvr_find(et131x);
		if(status) {
			printk(KERN_DEBUG "et131x: could not find xcvr\n");
		}
	}
	et131x_phy_init(et131x);
	et131x_phy_reset(et131x);
	et131x_phy_power(et131x, 0);
	et131x_phy_advertise_1000(et131x, 1, 1);
	et131x_phy_power(et131x, 1);
	et131x_setphy_normal(et131x);
	et131x_showregs(et131x);

	if((err = request_irq(pcidev->irq, et131x_intr, SA_SHIRQ, netdev->name, et131x))) {
		printk(KERN_DEBUG "et131x: could not request interrupt\n");
		goto exit_iounmap;
	}

	printk(KERN_DEBUG "et131x: PCI OK!\n");

	mdelay(5);

	if(!et131x_txdma_init(et131x))
		goto exit_free_irq;

	et131x_txdma_enable(et131x);
	et131x_txdma_send(et131x, testpacket, sizeof testpacket);

	// Fill in network device struct
	ether_setup(netdev);
	netdev->open = et131x_net_open;
	netdev->stop = et131x_net_close;
	netdev->hard_start_xmit = et131x_net_xmit;
	netdev->get_stats = et131x_net_stats;
	netdev->irq = pcidev->irq;
	netdev->base_addr = pcidev->resource[0].start;

	if((err = register_netdev(netdev))) {
		printk(KERN_ERR "et131x: unable to register network device\n");
		err = -ENOMEM;
		goto exit_free_irq;
	}
	
	et131x->regs->global.watchdog_timer = 650000000; // approx. 1 second
	et131x->regs->global.int_mask.watchdog = 0;

	return 0;

exit_dma_exit:
	et131x_txdma_disable(et131x);
	et131x_txdma_exit(et131x);
exit_free_irq:
	free_irq(pcidev->irq, et131x);
exit_iounmap:
	iounmap(et131x->regs);
exit_release_regions:
	pci_release_regions(pcidev);
exit_disable_device:
	pci_disable_device(pcidev);
exit_free_netdev:
	free_netdev(netdev);
exit:
	return err;
}

static void __devexit et131x_remove(struct pci_dev *pcidev) {
	struct et131x *et131x;

	printk(KERN_DEBUG "et131x: removing device\n");

	if(!pcidev)
		return;

	et131x = pci_get_drvdata(pcidev);

	if(!et131x) {
		printk(KERN_ERR "et131x: removal request for unknown device\n");
		return;
	}

	et131x_txdma_disable(et131x);
	et131x_txdma_exit(et131x);
	et131x->regs->global.int_mask.value = ~0;
	free_irq(pcidev->irq, et131x);
	iounmap(et131x->regs);
	unregister_netdev(et131x->netdev);
	free_netdev(et131x->netdev);
	pci_set_drvdata(pcidev, NULL);
	pci_release_regions(pcidev);
	pci_disable_device(pcidev);
}

static struct pci_device_id et131x_id_table[] = {
	{PCI_VENDOR_ID_ATT, 0xed00, PCI_ANY_ID, PCI_ANY_ID},
	{PCI_VENDOR_ID_ATT, 0xed01, PCI_ANY_ID, PCI_ANY_ID},
};

static struct pci_driver et131x_driver = {
	.name =     DRV_NAME,
	.id_table = et131x_id_table,
	.probe =    et131x_probe,
	.remove =   et131x_remove,
};

static int __init et131x_init(void) {
	printk(KERN_INFO "et131x: %s, %s\n", DRV_DESCRIPTION, DRV_VERSION);
	printk(KERN_INFO "et131x: %s\n", DRV_COPYRIGHT);
	return pci_module_init(&et131x_driver);
}

static void __exit et131x_exit(void) {
	printk(KERN_INFO "et131x: exiting\n");
	pci_unregister_driver(&et131x_driver);
}

module_init(et131x_init);
module_exit(et131x_exit);
MODULE_DESCRIPTION(DRV_DESCRIPTION);
MODULE_AUTHOR(DRV_COPYRIGHT);
MODULE_LICENSE(DRV_LICENSE);
