#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/pci.h>
#include <linux/etherdevice.h>

#include "et131x.h"

void et131x_wake(struct et131x *et131x) {
	union et131x_pm_ctrl pm_ctrl = {{
		.sysclk_gate = 1,
		.txclk_gate = 1,
		.rxclk_gate = 1,
	}};

	et131x->regs->global.pm_ctrl = pm_ctrl;
}

void et131x_soft_reset(struct et131x *et131x) {
	et131x->regs->mac.ctrl1.value = 0xc00f0000;
	wmb();
	et131x->regs->global.sw_reset.value = 0x7F;
	wmb();
	et131x->regs->mac.ctrl1.value = 0x000f0000;
	wmb();
	et131x->regs->mac.ctrl1.value = 0x00000000;
	wmb();
}

void et131x_disable_interrupts(struct et131x *et131x) {
	et131x->regs->global.int_mask.value = ~0;
	wmb();
}

void et131x_enable_interrupts(struct et131x *et131x) {
	et131x->regs->global.int_mask.value = 0xfff6bf17;
	wmb();
}

void et131x_config_global(struct et131x *et131x) {
	// Set up 1:1 Tx/Rx split
	et131x->regs->global.rxq_start_addr = 0;
	et131x->regs->global.rxq_end_addr = 0x1ff;
	et131x->regs->global.txq_start_addr = 0x200;
	et131x->regs->global.txq_end_addr = 0x3ff;

	// Disable stuff
	et131x->regs->global.loopback.value = 0;
	et131x->regs->global.msi_config.value = 0;
	et131x->regs->global.watchdog_timer = 0;

	wmb();
}

void et131x_config_mmc(struct et131x *et131x) {
	union et131x_mmc_ctrl ctrl = {{
		.enable = 1,
	}};
	
	et131x->regs->mmc.ctrl = ctrl;
	wmb();
}

void et131x_config_mac1(struct et131x *et131x) {
	static const union et131x_mac_ipg ipg = {{
		.non_b2b_1 = 0x38,
		.non_b2b_2 = 0x58,
		.min_enforce = 0x50,
		.b2b = 0x60,
	}};

	static const union et131x_mac_hfdp hfdp = {{
		.alt_beb_trunc = 0xA,
		.excess_defer = 0x1,
		.rexmit_max = 0xF,
		.coll_window = 0x37,
	}};

	static const union et131x_mac_mii_mgmt_cfg mgmt_cfg = {{
		.clk_reset = 0x7,
	}};

	et131x->regs->mac.ctrl1.value = 0xc00f0000;
	wmb();

	et131x->regs->mac.ipg = ipg;
	et131x->regs->mac.hfdp = hfdp;

	et131x->regs->mac.if_ctrl.value = 0;
	
	et131x->regs->mac.mii_mgmt_cfg = mgmt_cfg;

	// Copy MAC address from netdev
	et131x->regs->mac.station_addr2.octet1 = et131x->netdev->dev_addr[0];
	et131x->regs->mac.station_addr2.octet2 = et131x->netdev->dev_addr[1];
	et131x->regs->mac.station_addr1.octet3 = et131x->netdev->dev_addr[2];
	et131x->regs->mac.station_addr1.octet4 = et131x->netdev->dev_addr[3];
	et131x->regs->mac.station_addr1.octet5 = et131x->netdev->dev_addr[4];
	et131x->regs->mac.station_addr1.octet6 = et131x->netdev->dev_addr[5];

	et131x->regs->mac.max_fm_len = 9224; // Max jumbo packet size + 4 bytes CRC + 4 bytes VLAN

	wmb();
	et131x->regs->mac.ctrl1.value = 0;
	wmb();
}

void et131x_config_mac2(struct et131x *et131x) {
	// Read configuration
	union et131x_mac_ctrl1 ctrl1 = et131x->regs->mac.ctrl1;
	union et131x_mac_ctrl2 ctrl2 = et131x->regs->mac.ctrl2;
	union et131x_mac_if_ctrl if_ctrl = et131x->regs->mac.if_ctrl;
	union et131x_txmac_ctrl txmac_ctrl = et131x->regs->txmac.ctrl;

	// Gigabit (0x1, 0x1 for 100 Mbit)
	ctrl2.if_mode = 2;
	if_ctrl.phy_mode = 0;

	// Enable Rx/Tx
	ctrl1.rx_enable = 1;
	ctrl1.tx_enable = 1;

	// Enable flow control
	ctrl1.tx_flow = 1;
	ctrl1.rx_flow = 1;

	// Disable loopback
	ctrl1.loop_back = 0;

	ctrl2.preamble_len = 7;
	ctrl2.huge_frame = 0;
	ctrl2.len_check = 1;
	ctrl2.pad_crc = 1;
	ctrl2.crc_enable = 1;
	
	// Full duplex
	ctrl2.full_duplex = 1;
	if_ctrl.ghd_mode = !ctrl2.full_duplex;

	// Write configuration
	et131x->regs->mac.ctrl1 = ctrl1;
	et131x->regs->mac.if_ctrl = if_ctrl;
	et131x->regs->mac.ctrl2 = ctrl2;

	// TODO: check ctrl1.syncd_tx_en and _rx_en?
	
	wmb();
	txmac_ctrl.enable = 1;
	txmac_ctrl.fc_disable = 1;
	wmb();
	et131x->regs->txmac.ctrl = txmac_ctrl;
	
	// TODO: enable DMA
}

void et131x_config_rxmac(struct et131x *et131x) {
	// Disable MAC during configuration
	et131x->regs->rxmac.ctrl.value = 0;
	wmb();

	// Disable Wake on LAN
	et131x->regs->rxmac.crc0 = 0;
	et131x->regs->rxmac.crc1 = 0;
	et131x->regs->rxmac.crc2 = 0;
	et131x->regs->rxmac.crc3 = 0;
	et131x->regs->rxmac.crc4 = 0;

	et131x->regs->rxmac.wol_mask0_0 = 0;
	et131x->regs->rxmac.wol_mask0_1 = 0;
	et131x->regs->rxmac.wol_mask0_2 = 0;
	et131x->regs->rxmac.wol_mask0_3 = 0;
	et131x->regs->rxmac.wol_mask1_0 = 0;
	et131x->regs->rxmac.wol_mask1_1 = 0;
	et131x->regs->rxmac.wol_mask1_2 = 0;
	et131x->regs->rxmac.wol_mask1_3 = 0;
	et131x->regs->rxmac.wol_mask2_0 = 0;
	et131x->regs->rxmac.wol_mask2_1 = 0;
	et131x->regs->rxmac.wol_mask2_2 = 0;
	et131x->regs->rxmac.wol_mask2_3 = 0;
	et131x->regs->rxmac.wol_mask3_0 = 0;
	et131x->regs->rxmac.wol_mask3_1 = 0;
	et131x->regs->rxmac.wol_mask3_2 = 0;
	et131x->regs->rxmac.wol_mask3_3 = 0;

	et131x->regs->rxmac.wol_sa[0] = et131x->netdev->dev_addr[0];
	et131x->regs->rxmac.wol_sa[1] = et131x->netdev->dev_addr[1];
	et131x->regs->rxmac.wol_sa[2] = et131x->netdev->dev_addr[2];
	et131x->regs->rxmac.wol_sa[3] = et131x->netdev->dev_addr[3];
	et131x->regs->rxmac.wol_sa[4] = et131x->netdev->dev_addr[4];
	et131x->regs->rxmac.wol_sa[5] = et131x->netdev->dev_addr[5];

	// Disable packet filtering
	et131x->regs->rxmac.pf_ctrl.value = 0;

	// Initialise unicast packet filtering
	et131x->regs->rxmac.uni_pf_addr1_hi = 0;
	et131x->regs->rxmac.uni_pf_addr2_hi = 0;
	et131x->regs->rxmac.uni_pf_addr_lo.value = 0;

	// Disable multicast filtering
	// TODO: enable it properly?
	et131x->regs->rxmac.pf_ctrl.filter_multicast = 0;

	// TODO: Jumbo frames need different settings?
	et131x->regs->rxmac.mcif_ctrl_max_seg.value = 0;

	et131x->regs->rxmac.mcif_water_mark.value = 0;
	et131x->regs->rxmac.mif_ctrl.value = 0;
	et131x->regs->rxmac.space_avail.value = 0;
	// TODO: 100 Mbit needs different settings?
	et131x->regs->rxmac.mif_ctrl.value = 0x30030;
	et131x->regs->rxmac.pf_ctrl.min_pkt_size = 64;
	et131x->regs->rxmac.pf_ctrl.filter_fragments = 1;
	wmb();
	et131x->regs->rxmac.ctrl.value = 0x9;
	wmb();
}

void et131x_config_txmac(struct et131x *et131x) {
	et131x->regs->txmac.cf_param.cfpt = 0x40;
	et131x->regs->txmac.cf_param.cfep = 0x0;
	wmb();
}

void et131x_config_mac_stat(struct et131x *et131x) {
	// Ignore for now
}

void et131x_config_flowcontrol(struct et131x *et131x) {
	// TruePHY stuff	
}

