#include <linux/pci.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/delay.h>

#include "et131x.h"
#include "et131x_phy.h"
#include "et131x_phy_registers.h"

// TODO: convert this to use mii or libphy module.

int et131x_mdio_read(struct net_device *netdev, int addr, int reg) {
	uint32_t value;
	union et131x_mac_mii_mgmt_indicator status;
	struct et131x *et131x;
	int wait = 50;

	if(!netdev)
		return -ENODEV;

	et131x = netdev_priv(netdev);

	if(!et131x)
		return -ENODEV;
	

	writel(0, &et131x->regs->mac.mii_mgmt_cmd);
	writel((addr << 16) | reg, &et131x->regs->mac.mii_mgmt_addr);
	writel(1, &et131x->regs->mac.mii_mgmt_cmd);
	
	while(--wait) {
		udelay(50);
		status.value = readl(&et131x->regs->mac.mii_mgmt_indicator);
		if(!status.busy && !status.not_valid)
			break;
	}

	if(!wait) {
		printk(KERN_DEBUG "et131x: timeout while reading MII %d %d %08x\n", addr, reg, status.value);
		return -EIO;
	}
	
	value = readl(&et131x->regs->mac.mii_mgmt_stat);
	writel(0, &et131x->regs->mac.mii_mgmt_cmd);
	
	return value;
}	

void et131x_mdio_write(struct net_device *netdev, int addr, int reg, int value) {
	int wait = 50;
	union et131x_mac_mii_mgmt_indicator status;
	struct et131x *et131x;

	if(!netdev)
		return;

	et131x = netdev_priv(netdev);

	if(!et131x)
		return;
	
	writel(0, &et131x->regs->mac.mii_mgmt_cmd);
	writel((addr << 16) | reg, &et131x->regs->mac.mii_mgmt_addr);
	writel(value, &et131x->regs->mac.mii_mgmt_ctrl);
	
	while(--wait) {
		udelay(50);
		status.value = readl(&et131x->regs->mac.mii_mgmt_indicator);
		if(!status.busy && !status.not_valid)
			break;
	}

	if(!wait) {
		printk(KERN_DEBUG "et131x: timeout while writing MII %d %d %d %08x\n", addr, reg, value, status.value);
		return;
	}
	
	writel(0, &et131x->regs->mac.mii_mgmt_cmd);
}	

int et131x_xcvr_find(struct et131x *et131x) {
	union et131x_mii_idr1 idr1;
	union et131x_mii_idr2 idr2;
	int xcvr_addr;
	uint32_t xcvr_id;

	for (xcvr_addr = 0; xcvr_addr < 32; xcvr_addr++) {
		idr1.value = et131x_mdio_read(et131x->netdev, xcvr_addr, MII_PHYSID1);
		idr2.value = et131x_mdio_read(et131x->netdev, xcvr_addr, MII_PHYSID2);
		xcvr_id = (idr1.value << 16) | idr2.value;

		if(idr1.value != 0 && idr1.value != 0xffff) {
			printk(KERN_DEBUG "et131x: xcvr addr = %02x, id = %08x\n", xcvr_addr, xcvr_id);
			et131x->xcvr_id = xcvr_id;
			et131x->xcvr_addr = xcvr_addr;
			return 0;
		}
	}

	printk(KERN_ERR "et131x: no xcvr found\n");
	return -ENODEV;
}

void et131x_phy_reset(struct et131x *et131x) {
	et131x_mdio_write(et131x->netdev, et131x->xcvr_addr, MII_BMCR, 0x8000);
}

void et131x_phy_power(struct et131x *et131x, int enable) {
	union et131x_mii_bmcr bmcr;

	bmcr.value = et131x_mdio_read(et131x->netdev, et131x->xcvr_addr, MII_BMCR);
	bmcr.power_down = !enable;
	et131x_mdio_write(et131x->netdev, et131x->xcvr_addr, MII_BMCR, bmcr.value);
}

void et131x_phy_autoneg(struct et131x *et131x, int enable) {
	union et131x_mii_bmcr bmcr;

	bmcr.value = et131x_mdio_read(et131x->netdev, et131x->xcvr_addr, MII_BMCR);
	bmcr.enable_autoneg = enable;
	et131x_mdio_write(et131x->netdev, et131x->xcvr_addr, MII_BMCR, bmcr.value);
}

void et131x_phy_duplex_mode(struct et131x *et131x, int fullduplex) {
	union et131x_mii_bmcr bmcr;

	bmcr.value = et131x_mdio_read(et131x->netdev, et131x->xcvr_addr, MII_BMCR);
	bmcr.duplex_mode = fullduplex;
	et131x_mdio_write(et131x->netdev, et131x->xcvr_addr, MII_BMCR, bmcr.value);
}

void et131x_phy_speed(struct et131x *et131x, int speed) {
	union et131x_mii_bmcr bmcr;

	bmcr.value = et131x_mdio_read(et131x->netdev, et131x->xcvr_addr, MII_BMCR);
	bmcr.speed_1000_sel = speed == 1000;
	bmcr.speed_sel = speed == 100;
	et131x_mdio_write(et131x->netdev, et131x->xcvr_addr, MII_BMCR, bmcr.value);
}

void et131x_phy_advertise_1000(struct et131x *et131x, int halfduplex, int fullduplex) {
	uint16_t ctrl1000;

	ctrl1000 = et131x_mdio_read(et131x->netdev, et131x->xcvr_addr, MII_CTRL1000);
	ctrl1000 &= ~0x0300;

	if(halfduplex)
		ctrl1000 |= 0x100;
	if(fullduplex)
		ctrl1000 |= 0x200;

	et131x_mdio_write(et131x->netdev, et131x->xcvr_addr, MII_CTRL1000, ctrl1000);
}


void et131x_phy_advertise_100(struct et131x *et131x, int halfduplex, int fullduplex) {
	uint16_t advertise;

	advertise = et131x_mdio_read(et131x->netdev, et131x->xcvr_addr, MII_ADVERTISE);
	advertise &= ~0x0180;

	if(halfduplex)
		advertise &= 0x0080;
	if(fullduplex)
		advertise &= 0x0100;

	et131x_mdio_write(et131x->netdev, et131x->xcvr_addr, MII_ADVERTISE, advertise);
}

void et131x_phy_advertise_10(struct et131x *et131x, int halfduplex, int fullduplex) {
	uint16_t advertise;

	advertise = et131x_mdio_read(et131x->netdev, et131x->xcvr_addr, MII_ADVERTISE);
	advertise &= ~0x0060;

	if(halfduplex)
		advertise &= 0x0020;
	if(fullduplex)
		advertise &= 0x0040;

	et131x_mdio_write(et131x->netdev, et131x->xcvr_addr, MII_ADVERTISE, advertise);
}

void et131x_phy_linkstatus(struct et131x *et131x, int *linkstatus, int *autoneg, int *linkspeed, int *duplex, int *mdimix, int *master, int *polarity) {
	uint16_t bmcr, bmsr, stat1000, psr;

	bmcr = et131x_mdio_read(et131x->netdev, et131x->xcvr_addr, MII_BMCR);
	bmsr = et131x_mdio_read(et131x->netdev, et131x->xcvr_addr, MII_BMSR);
	stat1000 = et131x_mdio_read(et131x->netdev, et131x->xcvr_addr, MII_STAT1000);
	psr = et131x_mdio_read(et131x->netdev, et131x->xcvr_addr, ET131X_MII_PSR);

	if(linkstatus)
		*linkstatus = psr & 0x0040;
	if(autoneg)
		*autoneg = (bmcr & 0x1000) ? ((psr & 0x0020) ? 1 : 0) : -1;
	if(linkspeed)
		*linkspeed = (psr & 0x0200) ? ((psr & 0x100) ? 1000 : 100) : ((psr & 0x0100) ? 10 : 0);
	if(duplex)
		*duplex = psr & 0x0080;
	if(mdimix)
		*mdimix = 0;
	if(master)
		*master = stat1000 & 0x4000;
	if(polarity)
		*polarity = psr & 0x0400;
}

int et131x_xcvr_init(struct et131x *et131x) {
	union et131x_mii_bmcr bmcr;
	union et131x_mii_imr imr;
	union et131x_mii_isr isr;
	union et131x_mii_anar anar;
	union et131x_mii_lcr2 lcr2;

	isr.value = et131x_mdio_read(et131x->netdev, et131x->xcvr_addr, ET131X_MII_ISR);
	imr.value = et131x_mdio_read(et131x->netdev, et131x->xcvr_addr, ET131X_MII_IMR);
	imr.int_en = 1;
	imr.link_status = 1;
	imr.autoneg_status = 1;
	et131x_mdio_write(et131x->netdev, et131x->xcvr_addr, ET131X_MII_IMR, imr.value);

	lcr2.value = et131x_mdio_read(et131x->netdev, et131x->xcvr_addr, ET131X_MII_LCR2);
	lcr2.led_tx_rx = 0x3;
	lcr2.led_link = 0xa;
	et131x_mdio_write(et131x->netdev, et131x->xcvr_addr, ET131X_MII_LCR, lcr2.value);

	anar.value = et131x_mdio_read(et131x->netdev, et131x->xcvr_addr, MII_ADVERTISE);
	anar.cap_pause = 1;
	anar.cap_asmpause = 1;
	et131x_mdio_write(et131x->netdev, et131x->xcvr_addr, MII_ADVERTISE, anar.value);

	et131x_phy_autoneg(et131x, 1);

	bmcr.value = et131x_mdio_read(et131x->netdev, et131x->xcvr_addr, MII_BMCR);
	bmcr.restart_autoneg = 1;
	et131x_mdio_write(et131x->netdev, et131x->xcvr_addr, MII_BMCR, bmcr.value);

	return 0;
}

static uint16_t configphy[25][2] = {
	/* Reg      Value      Register */
	/* Addr                         */
	{0x880B, 0x0926},       /* AfeIfCreg4B1000Msbs */
	{0x880C, 0x0926},       /* AfeIfCreg4B100Msbs */
	{0x880D, 0x0926},       /* AfeIfCreg4B10Msbs */

	{0x880E, 0xB4D3},       /* AfeIfCreg4B1000Lsbs */
	{0x880F, 0xB4D3},       /* AfeIfCreg4B100Lsbs */
	{0x8810, 0xB4D3},       /* AfeIfCreg4B10Lsbs */

	{0x8805, 0xB03E},       /* AfeIfCreg3B1000Msbs */
	{0x8806, 0xB03E},       /* AfeIfCreg3B100Msbs */
	{0x8807, 0xFF00},       /* AfeIfCreg3B10Msbs */

	{0x8808, 0xE090},       /* AfeIfCreg3B1000Lsbs */
	{0x8809, 0xE110},       /* AfeIfCreg3B100Lsbs */
	{0x880A, 0x0000},       /* AfeIfCreg3B10Lsbs */

 	{0x300D, 1},            /* DisableNorm */

 	{0x280C, 0x0180},       /* LinkHoldEnd */

 	{0x1C21, 0x0002},       /* AlphaM */

	{0x3821, 6},            /* FfeLkgTx0 */
	{0x381D, 1},            /* FfeLkg1g4 */
	{0x381E, 1},            /* FfeLkg1g5 */
	{0x381F, 1},            /* FfeLkg1g6 */
	{0x3820, 1},            /* FfeLkg1g7 */

	{0x8402, 0x01F0},       /* Btinact */
	{0x800E, 20},           /* LftrainTime */
	{0x800F, 24},           /* DvguardTime */
	{0x8010, 46},           /* IdlguardTime */

	{0, 0}
};

void et131x_phy_init(struct et131x *et131x) {
	union et131x_mii_idr1 idr1;
	union et131x_mii_idr2 idr2;
	int mcr, data, bmcr;
	int i;

	// get the identity (again ?)
	idr1.value = et131x_mdio_read(et131x->netdev, et131x->xcvr_addr, MII_PHYSID1);
	idr2.value = et131x_mdio_read(et131x->netdev, et131x->xcvr_addr, MII_PHYSID2);

	// what does this do/achieve ?
	mcr = et131x_mdio_read(et131x->netdev, et131x->xcvr_addr, ET131X_MII_MCR);	// should read 0002
	et131x_mdio_write(et131x->netdev, et131x->xcvr_addr, ET131X_MII_MCR, 0x0006);

	// read modem register 0402, should I do something with the return data ?
	et131x_mdio_write(et131x->netdev, et131x->xcvr_addr, ET131X_MII_INDEX, 0x0402);
	data = et131x_mdio_read(et131x->netdev, et131x->xcvr_addr, ET131X_MII_DATA);

	// what does this do/achieve ?    
	et131x_mdio_write(et131x->netdev, et131x->xcvr_addr, ET131X_MII_MCR, 0x0002);

	// get the identity (again ?)
	idr1.value = et131x_mdio_read(et131x->netdev, et131x->xcvr_addr, MII_PHYSID1);
	idr2.value = et131x_mdio_read(et131x->netdev, et131x->xcvr_addr, MII_PHYSID2);

	// what does this achieve ?    
	mcr = et131x_mdio_read(et131x->netdev, et131x->xcvr_addr, ET131X_MII_MCR);	// should read 0002
	et131x_mdio_write(et131x->netdev, et131x->xcvr_addr, ET131X_MII_MCR, 0x0006);

	// read modem register 0402, should I do something with the return data ?
	et131x_mdio_write(et131x->netdev, et131x->xcvr_addr, ET131X_MII_INDEX, 0x0402);
	data = et131x_mdio_read(et131x->netdev, et131x->xcvr_addr, ET131X_MII_DATA);
	et131x_mdio_write(et131x->netdev, et131x->xcvr_addr, ET131X_MII_MCR, 0x0002);

	// what does this achieve (should return 0x1040)
	bmcr = et131x_mdio_read(et131x->netdev, et131x->xcvr_addr, MII_BMCR);
	mcr = et131x_mdio_read(et131x->netdev, et131x->xcvr_addr, ET131X_MII_MCR);	// should read 0002
	et131x_mdio_write(et131x->netdev, et131x->xcvr_addr, MII_BMCR, 0x1840);
	et131x_mdio_write(et131x->netdev, et131x->xcvr_addr, ET131X_MII_MCR, 0x0007);

	// write configphy array
	for(i = 0; configphy[i][0]; i++) {
		// write value
		et131x_mdio_write(et131x->netdev, et131x->xcvr_addr, ET131X_MII_INDEX, configphy[i][0]);
		et131x_mdio_write(et131x->netdev, et131x->xcvr_addr, ET131X_MII_DATA, configphy[i][1]);

		// read it back
		et131x_mdio_write(et131x->netdev, et131x->xcvr_addr, ET131X_MII_INDEX, configphy[i][0]);
		data = et131x_mdio_read(et131x->netdev, et131x->xcvr_addr, ET131X_MII_DATA);

		// do a check on the value read back ?
	}

	bmcr = et131x_mdio_read(et131x->netdev, et131x->xcvr_addr, MII_BMCR);	// 0x1840
	mcr = et131x_mdio_read(et131x->netdev, et131x->xcvr_addr, ET131X_MII_MCR);	// should read 0007
	et131x_mdio_write(et131x->netdev, et131x->xcvr_addr, MII_BMCR, 0x1040);
	et131x_mdio_write(et131x->netdev, et131x->xcvr_addr, ET131X_MII_MCR, 0x0002);
}

int et131x_setphy_normal(struct et131x *et131x) {
	et131x_phy_power(et131x, 1);
	return et131x_xcvr_init(et131x);
}


