//****************************************************************************
// Copyright (C) 2006  PEAK System-Technik GmbH
//
// linux@peak-system.com
// www.peak-system.com
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// Maintainer(s): Klaus Hitschler (klaus.hitschler@gmx.de)
//****************************************************************************

//***************************************************************************
//
// all parts to handle the interface specific parts of pcan-pccard
//
// $Id: pcan_pccard.c 367 2006-03-26 16:56:55Z khitschler $
//
//****************************************************************************

//****************************************************************************
// INCLUDES
#include <src/pcan_common.h>     // must always be the 1st include
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/delay.h>

#include <pcmcia/cs_types.h>
#include <pcmcia/cs.h>
#include <pcmcia/cistpl.h>
#include <pcmcia/ciscode.h>
#include <pcmcia/ds.h>
#include <pcmcia/cisreg.h>
#include <asm/io.h>
#include <linux/spinlock.h>

#include <src/pcan_main.h>
#include <src/pcan_pccard.h>
#include <src/pcan_sja1000.h>
#include <src/pcan_fifo.h>

//****************************************************************************
// DEFINES
#define PCCARD_MINOR_BASE  40           // starting point of minors for PCCARD devices

#define PCCARD_MANF_ID     0x0377       // manufacturer PEAK System GmbH         
#define PCCARD_CARD_ID     0x0001       // card id
#define PCCARD_CHANNELS    2            // maximum n channels per card

#define PCCARD_PORT_SIZE   0x20         // range of a channels port
#define PCCARD_COMMON_SIZE 0x40         // range of channels common ports
#define PCCARD_CHANNEL_OFF 0x20         // port offset of sja1000 channel #2
#define PCCARD_COMMON      0x40         // port offset of common area

                                        // offsets from base of common area
#define CCR                0x00         // CAN control register
#define CSR                0x02         // CAN status register
#define CPR                0x04         // CAN power register
#define SPIDATI            0x06         // SPI data in
#define SPIDATO            0x08         // SPI data out
#define SPIADR             0x0A         // SPI address register
#define SPIINS             0x0C         // SPI instruction register
#define FW_MAJOR           0x10         // firmware major number (local)
#define FW_MINOR           0x12         // firmware minor number (local)

#define CCR_CLK_MASK       0x03         // mask for clock code
#define CCR_CLK16          0x00         // clk code
#define CCR_CLK10          0x01         // 
#define CCR_CLK21          0x02         // 
#define CCR_CLK8           0x03         // 

#define CCR_RESET_MASK     0x01
#define CCR_RESET          0x01         // put channel into reset
#define CCR_RESET0_SHIFT      2         // shift for channel 0
#define CCR_RESET1_SHIFT      3         // shift for channel 1

#define CCR_LED_MASK       0x03         // mask for LED channel 0 code
#define CCR_LED_ON         0x00         // switch LED permanently on
#define CCR_LED_FAST       0x01         // switch LED to 4 Hz flash
#define CCR_LED_SLOW       0x02         // switch LED to 1 Hz flash
#define CCR_LED_OFF        0x03         // switch LED off
#define CCR_LED0_SHIFT        4         // shift for channel 1
#define CCR_LED1_SHIFT        6         // shift for channel 1

#define CCR_DEFAULT       ((((u8)CCR_LED_OFF) << CCR_LED0_SHIFT) | (((u8)CCR_LED_OFF) << CCR_LED1_SHIFT)) 

#define CSR_SPIBUSY        0x04         // set when SPI transaction is busy

#define CCR_CLK_DEFAULT    CCR_CLK16    // set default clock       
#define SPI_MAX_WAIT_CYCLES 100

//****************************************************************************
// helper macros
#define CS_CHECK(fn, ret) do { last_fn = (fn); if ((last_ret = (ret)) != 0) goto fail; } while (0)

#define CS_PREPARE(tpl) { tuple.DesiredTuple = tpl;  tuple.TupleData = buf; tuple.TupleOffset = 0; \
                          tuple.TupleDataMax = CISTPL_END; tuple.Attributes   = 0; }
  
#define WAIT_WHILE_SPI_ENGINE_BUSY() \
{\
  counter = SPI_MAX_WAIT_CYCLES;\
  do\
    schedule();\
  while ((pccard_readreg_common(card, CSR) & CSR_SPIBUSY) && counter--);\
  if (counter <= 0)\
  {\
    err = -EIO;\
    goto fail;\
  }\
}
  
//****************************************************************************
// GLOBALS

//****************************************************************************
// LOCALS
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
static dev_link_t *pccard_card_list = NULL;
static dev_info_t  pccard_info      = DEVICE_NAME;
#endif

static u16 pccard_devices = 0;

typedef struct pcan_pccard
{
  dev_link_t link;
  u32        dwPort;                         // channels commonly used port 
  spinlock_t lock;                           // locking access registers accessed from both channels
  struct     pcandev *dev[PCCARD_CHANNELS];  // point to associated channels
  dev_node_t node;                           // to satisfy cardmgr needs
} PCAN_PCCARD;

//****************************************************************************
// CODE

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,13)
static struct pcmcia_device_id pccard_id_table[] =
{
  PCMCIA_DEVICE_MANF_CARD(PCCARD_MANF_ID, PCCARD_CARD_ID),
  PCMCIA_DEVICE_NULL,
};
#endif

//****************************************************************************
// read channel individual register
static u8 pccard_readreg(struct pcandev *dev, u8 port) // read a register
{
  return inb(dev->port.pccard.dwPort + port);
}

// write channel individual register
static void pccard_writereg(struct pcandev *dev, u8 port, u8 data) // write a register
{
  outb(data, dev->port.pccard.dwPort + port) ; 
}

// read common register to both channels of a card
static inline u8 pccard_readreg_common(PCAN_PCCARD *card, u8 port) // read a register
{
  return inb(card->dwPort + port);
}

// write common register to both channels of a card
static inline void pccard_writereg_common(PCAN_PCCARD *card, u8 port, u8 data) // write a register
{
  outb(data, card->dwPort + port); 
}

//****************************************************************************
// SPI engine
static int pccard_write_eeprom(PCAN_PCCARD *card, u16 addr, u8 val)
{
  u16 status;
  int err = 0;
  int counter;

  // Instruction WRITE
  pccard_writereg_common(card, SPIINS, 0x06); // WREN
  WAIT_WHILE_SPI_ENGINE_BUSY();
    
  // warte, bis WEL gesetzt ist
  do 
  {
    pccard_writereg_common(card, SPIINS, 0x05); // RDSR == Read status
    WAIT_WHILE_SPI_ENGINE_BUSY();    
    status = pccard_readreg_common(card, SPIDATI);
  } while (!(status & 0x02)) ; // warte, bis WEL (write enable) 1.

  // Adresse und Data setzen
  pccard_writereg_common(card, SPIADR, addr & 0xff);
  pccard_writereg_common(card, SPIDATO, val);
  
  // Instruction WRITE
  pccard_writereg_common(card, SPIINS, ((addr & 0x100) ? 8 : 0) | 0x02); // WRITE mit bit3 = Addr8
  WAIT_WHILE_SPI_ENGINE_BUSY();

  // warte, bis Schreiben abgeschlossen
  do 
  {
      pccard_writereg_common(card, SPIINS, 0x05); // RDSR == Read status
      WAIT_WHILE_SPI_ENGINE_BUSY();
      status = pccard_readreg_common(card, SPIDATI);
  } while (status & 0x01) ; // warte, bis WIP (Write In Progress) 0.
  
  fail:
  if (err)
  {
    DPRINTK(KERN_DEBUG "%s: pccard_write_eeprom(0x%04x, 0x%02x) failed!\n", DEVICE_NAME, addr, val);
  }
  return err;
}

#if 0
static int pccard_read_eeprom(PCAN_PCCARD *card, u16 addr, u8 *val)
{
  int err = 0;
  int counter;
  
  pccard_writereg_common(card, SPIADR, addr & 0xff);
  pccard_writereg_common(card, SPIINS, ((addr & 0x100) ? 8 : 0) | 0x03); // READ mit bit3 = Addr8
  WAIT_WHILE_SPI_ENGINE_BUSY();
  *val = pccard_readreg_common(card, SPIDATI);
  
  fail:
  if (err)
  {
    DPRINTK(KERN_DEBUG "%s: pccard_read_eeprom(0x%04x) failed!\n", DEVICE_NAME, addr);
  }
  return err;
}
#endif

//****************************************************************************
// helper functions

// get firmware number of PCC
static inline void pccard_show_firmware_version(PCAN_PCCARD *card)
{
  u8 fw_major = pccard_readreg_common(card, FW_MAJOR);
  u8 fw_minor = pccard_readreg_common(card, FW_MINOR);
  
  printk(KERN_INFO "%s: pccard firmware %d.%d\n", DEVICE_NAME, fw_major, fw_minor);
}

// init CCR settings
static inline void pccard_initreg_common(PCAN_PCCARD *card)
{
  spin_lock_init(&card->lock);
  spin_lock(&card->lock);
  pccard_writereg_common(card, CCR, CCR_DEFAULT);  
  spin_unlock(&card->lock);
}

// hard reset only one channel to its default settings
static inline void pccard_channel_reset(PCAN_PCCARD *card, int nChannel)
{
  u8 data;
  u8 shift = (nChannel) ? CCR_RESET1_SHIFT : CCR_RESET0_SHIFT;
  
  DPRINTK(KERN_DEBUG "%s: pccard_channel_reset(%d)\n", DEVICE_NAME, nChannel);
      
  spin_lock(&card->lock);
  data = pccard_readreg_common(card, CCR);
  data &= ~(CCR_RESET_MASK << shift);
  data |=  (CCR_RESET      << shift);
  pccard_writereg_common(card, CCR, data);
  spin_unlock(&card->lock);
  mdelay(2);

  spin_lock(&card->lock);
  data = pccard_readreg_common(card, CCR);
  data &= ~(CCR_RESET_MASK << shift);
  data &= ~(CCR_RESET      << shift);
  pccard_writereg_common(card, CCR, data);  
  spin_unlock(&card->lock);
  mdelay(10); // wait until reset has settled
}

// set LED
static inline void pccard_set_LED(PCAN_PCCARD *card, int nChannel, u8 mode)
{
  u8 data;
  u8 shift = (nChannel) ? CCR_LED1_SHIFT : CCR_LED0_SHIFT;
  
  spin_lock(&card->lock);
  data = pccard_readreg_common(card, CCR);
  data &= ~(CCR_LED_MASK         << shift);
  data |= ((CCR_LED_MASK & mode) << shift);
  pccard_writereg_common(card, CCR, data);
  spin_unlock(&card->lock);
}

// enable CAN power
static inline void pccard_enable_CAN_power(PCAN_PCCARD *card)
{
  pccard_write_eeprom(card, 0, 1);
}

// disable CAN power
static inline void pccard_disable_CAN_power(PCAN_PCCARD *card)
{
  pccard_write_eeprom(card, 0, 0);
}

//****************************************************************************
// all about interrupt handling
irqreturn_t pccard_irqhandler(int irq, void *dev_id, struct pt_regs *regs)
{
  irqreturn_t ret = 0;
  struct pcandev *dev = (struct pcandev *)dev_id;
  u8 fastBlink = (pcan_fifo_empty(&dev->writeFifo)) ? CCR_LED_FAST : CCR_LED_SLOW;
  
  #if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,24)
  sja1000_base_irqhandler(irq, dev_id, regs);
  #else
  ret = sja1000_base_irqhandler(irq, dev_id, regs);
  #endif
 
  fastBlink = (pcan_fifo_empty(&dev->readFifo)) ? CCR_LED_FAST : fastBlink;
  
  pccard_set_LED(dev->port.pccard.card, dev->port.pccard.nChannel, fastBlink);
     
  return ret;
}

static int pccard_req_irq(struct pcandev *dev)
{
  int err;
  
  DPRINTK(KERN_DEBUG "%s: pccard_req_irq()\n", DEVICE_NAME);
  
  if (dev->wInitStep == 4)
  {
    if ((err = request_irq(dev->port.pccard.wIrq, pccard_irqhandler, SA_INTERRUPT | SA_SHIRQ, "pcan", dev)))
      return err;
        
    dev->wInitStep = 5;
  }
  
  return 0;
}

static void pccard_free_irq(struct pcandev *dev)
{
  DPRINTK(KERN_DEBUG "%s: pccard_free_irq()\n", DEVICE_NAME);
  
  if (dev->wInitStep >= 5)
  {  
    free_irq(dev->port.pccard.wIrq, dev);  
    dev->wInitStep = 4;
  }
}

//****************************************************************************
// free allocated resources
static int pccard_cleanup(struct pcandev *dev)
{
  DPRINTK(KERN_DEBUG "%s: pccard_cleanup()\n", DEVICE_NAME);
  
  if (dev)
  {
    PCAN_PCCARD *card = dev->port.pccard.card;
    
    switch(dev->wInitStep)
    {
      case 5: pccard_free_irq(dev); 
      case 4: if (card) 
                pccard_set_LED(card, dev->port.pccard.nChannel, CCR_LED_OFF); 
      case 3: pcan_drv.wDeviceCount--;
              pccard_devices--; 
      case 2: dev->ucPhysicallyInstalled = 0;
              list_del(&dev->list);    
      case 1: 
              #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,8)
              DPRINTK(KERN_DEBUG "%s: release_region(PCCARD_PORT_SIZE)\n", DEVICE_NAME);
              release_region(dev->port.pccard.dwPort, PCCARD_PORT_SIZE);
              #endif
      case 0: dev->wInitStep = 0;
              DPRINTK(KERN_DEBUG "%s: kfree(0x%p)\n", DEVICE_NAME, dev);
              kfree(dev); 
    }
  } 
  return 0;
}

//****************************************************************************
// interface depended open and close
static int pccard_open(struct pcandev *dev)
{
  DPRINTK(KERN_DEBUG "%s: pccard_open()\n", DEVICE_NAME);
  pccard_set_LED(dev->port.pccard.card, dev->port.pccard.nChannel, CCR_LED_SLOW);
  dev->port.pccard.card->link.open++;
  return 0;
}

static int pccard_release(struct pcandev *dev)
{
  DPRINTK(KERN_DEBUG "%s: pccard_release()\n", DEVICE_NAME);
  pccard_set_LED(dev->port.pccard.card, dev->port.pccard.nChannel, CCR_LED_ON);
  if (dev->port.pccard.card->link.open > 0)
    dev->port.pccard.card->link.open--;
  return 0;
}

//****************************************************************************
// create and init a single pccard pcan device structure
static struct pcandev *pccard_create_single_device(PCAN_PCCARD *card, int nChannel, u32 dwPort, u16  wIrq)
{
  struct pcandev *dev;
  PCCARD_PORT    *p;
  int err = 0;
  
  DPRINTK(KERN_DEBUG "%s: pccard_create_single_device(%d, 0x%04x, %d)\n", DEVICE_NAME, nChannel, dwPort, wIrq);
  
  // allocate memory for my device
  if ((dev = (struct pcandev *)kmalloc(sizeof(struct pcandev), GFP_ATOMIC)) == NULL)
  {
    printk(KERN_ERR "%s: pccard_create_single_device - memory allocation failed!\n", DEVICE_NAME);
    err = -ENOMEM;
    goto fail;
  }
  DPRINTK(KERN_DEBUG "%s: kmalloc() = 0x%p\n", DEVICE_NAME, dev);
  memset(dev, 0x00, sizeof(*dev));

  dev->wInitStep = 0;

  // init structure elements to defaults
  pcan_soft_init(dev, "pccard", HW_PCCARD);

  // device handling standards for sja1000 driven parts
  dev->device_open      = sja1000_open;
  dev->device_release   = sja1000_release;
  dev->device_write     = sja1000_write;
  
  // init process wait queues
  init_waitqueue_head(&dev->read_queue);
  init_waitqueue_head(&dev->write_queue);
    
  // set this before any instructions, fill struct pcandev, part 1 
  dev->readreg     = pccard_readreg;
  dev->writereg    = pccard_writereg;
  dev->cleanup     = pccard_cleanup;
  dev->req_irq     = pccard_req_irq;
  dev->free_irq    = pccard_free_irq;
  dev->open        = pccard_open;
  dev->release     = pccard_release;
  dev->nMinor      = PCCARD_MINOR_BASE + pccard_devices; 
  
  dev->props.ucExternalClock = 1;
  dev->props.ucMasterDevice  = (nChannel) ? CHANNEL_SLAVE : CHANNEL_MASTER;
           
  // reject illegal combination
  if (!dwPort || !wIrq)
  {
    err = -EINVAL;
    goto fail;
  }
  
  // fill struct pcandev, hardware specfic parts
  p           = &dev->port.pccard;
  p->dwPort   = dwPort;
  p->wIrq     = wIrq;
  p->nChannel = nChannel;
  p->card     = card;

  #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,8)
  // allocating io resources
  DPRINTK(KERN_DEBUG "%s: ___request_region(PCCARD_PORT_SIZE)\n", DEVICE_NAME);
  err = ___request_region(dev->port.pccard.dwPort, PCCARD_PORT_SIZE, DEVICE_NAME);
  if (err)
    goto fail;
  #endif
  dev->wInitStep = 1;

  // base init of hardware - reset channel
  pccard_channel_reset(card, nChannel); 
  
  // probe hardware
  if ((err = sja1000_probe(dev)))
    goto fail;  
  
  // add into list of devices and assign the device as plugged in
  dev->ucPhysicallyInstalled = 1;
  list_add_tail(&dev->list, &pcan_drv.devices);  // add this device to the list       
  dev->wInitStep = 2;
  
  // increment device counts
  pcan_drv.wDeviceCount++;
  pccard_devices++; 
  dev->wInitStep = 3;
    
  pccard_set_LED(card, nChannel, CCR_LED_ON);
  dev->wInitStep = 4;
  
  printk(KERN_INFO "%s: pccard device minor %d found\n", DEVICE_NAME, dev->nMinor);

  return dev;

  fail:
  pccard_cleanup(dev);
  DPRINTK(KERN_INFO "%s: pccard_create_single_device(%d) failed! (%d)\n", DEVICE_NAME, nChannel, err);
  
  return NULL;
}

//****************************************************************************
// allocate resources and common io area and initialize hardware of devices
static int pccard_create_all_devices(dev_link_t *link)
{
  int err = 0;
  PCAN_PCCARD *card = link->priv;
  
  DPRINTK(KERN_DEBUG "%s: pccard_create_all_devices()\n", DEVICE_NAME);
    
  if (card)
  {
    ioaddr_t base = link->io.BasePort1;
    u_int    irq  = link->irq.AssignedIRQ;
    int      chn  = 0;
    struct pcandev *dev;
    int i;
    
    // fill card structure
    card->dwPort = base + PCCARD_COMMON;
  
    #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,8)
    // request commonly used io area
    
    DPRINTK(KERN_DEBUG "%s: ___request_region(PCCARD_COMMON_SIZE)\n", DEVICE_NAME);
    err = ___request_region(card->dwPort, PCCARD_COMMON_SIZE, DEVICE_NAME);
    if (err)
      goto fail;
    #endif
      
    // init common registers
    pccard_initreg_common(card);
      
    for (i = 0; i < PCCARD_CHANNELS; i++)
    {
      card->dev[i] = NULL;
      
      // probe devices and if successfully create device structues - channel 0
      dev = pccard_create_single_device(card, i, base + i * PCCARD_CHANNEL_OFF, irq);
      if (dev) 
      {
        chn++;
        card->dev[i] = dev;
      }
    }
    
    // no usable channel found
    if (!chn)
    {
      printk(KERN_ERR "%s: pccard with no channels found!\n", DEVICE_NAME);
      err = -ENODEV;
      goto fail;
    }
    
    return 0;
  }
  
  fail:
  DPRINTK(KERN_DEBUG "%s: pccard_create_all_devices(%d) failed!\n", DEVICE_NAME, err);
  return err;
}

//****************************************************************************
// deinit devices and release allocted resources of devices and common io area
static void pccard_release_all_devices(dev_link_t *link)
{
  PCAN_PCCARD *card = link->priv;
  
  DPRINTK(KERN_DEBUG "%s: pccard_release_all_devices()\n", DEVICE_NAME);
  
  if (card)
  {
    struct pcandev *dev;
    int i;
    
    for (i = 0; i < PCCARD_CHANNELS; i++)
    {
      // 1. release associated devices
      dev = card->dev[i];
      pccard_cleanup(dev);
      card->dev[i] = NULL;
    }
    
    #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,8)
    DPRINTK(KERN_DEBUG "%s: release_region(PCCARD_COMMON_SIZE)\n", DEVICE_NAME);
    release_region(card->dwPort, PCCARD_COMMON_SIZE);
    #endif
  }
}

//****************************************************************************
// is called at CARD_REMOVAL
static int pccard_plugout(dev_link_t *link)
{
  DPRINTK(KERN_DEBUG "%s: pccard_plugout()\n", DEVICE_NAME);
  
  #if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
  if (link->open) 
  {
    PCAN_PCCARD *card = link->priv;
    int i;
    
    printk(KERN_INFO "%s: release postponed since device is still used!\n", DEVICE_NAME);
    link->state |= DEV_STALE_CONFIG;
    
    // mark all associated channels as plugged out
    for (i = 0; i < PCCARD_CHANNELS; i++)
      if (card->dev[i])
        card->dev[i]->ucPhysicallyInstalled = 0;
         
    return -EBUSY;
  }
  #endif

  link->state &= ~DEV_PRESENT;
  
  if (link->state & DEV_CONFIG)
  {    
    // diable CAN power
    pccard_disable_CAN_power(link->priv);
    
    pccard_release_all_devices(link);
        
    // release pcmcia driver part
    link->dev  = NULL;
    pcmcia_release_configuration(link->handle);
    pcmcia_release_io(link->handle, &link->io);
    pcmcia_release_irq(link->handle, &link->irq);
    
    link->state &= ~DEV_CONFIG;
  }
  
  DPRINTK(KERN_DEBUG "%s: pccard_plugout() is OK\n", DEVICE_NAME);
  return 0;
}

//****************************************************************************
// is called at CARD_INSERTION or startup with still inserted card
static int pccard_plugin(dev_link_t *link)
{
  client_handle_t handle = link->handle;
  tuple_t   tuple;
  cisdata_t buf[CISTPL_END + 1];
  cisparse_t parse;
  config_info_t conf;
  cistpl_vers_1_t *cv;
  int last_ret;
  int last_fn;
  PCAN_PCCARD *card = NULL;
  
  DPRINTK(KERN_DEBUG "%s: pccard_plugin()\n", DEVICE_NAME);
  
  // show manufacturer string in log
  CS_PREPARE(CISTPL_VERS_1);
  CS_CHECK(GetFirstTuple, pcmcia_get_first_tuple(handle, &tuple));
  CS_CHECK(GetTupleData, pcmcia_get_tuple_data(handle, &tuple));
  CS_CHECK(ParseTuple, pcmcia_parse_tuple(handle, &tuple, &parse));
  cv = &parse.version_1;
  printk(KERN_INFO "%s: %s %s %s %s\n", DEVICE_NAME, &cv->str[cv->ofs[0]], &cv->str[cv->ofs[1]], 
                                                     &cv->str[cv->ofs[2]], &cv->str[cv->ofs[3]]);

  // specify what to request
  CS_PREPARE(CISTPL_CONFIG);
  CS_CHECK(GetFirstTuple, pcmcia_get_first_tuple(handle, &tuple));
  CS_CHECK(GetTupleData, pcmcia_get_tuple_data(handle, &tuple));
  CS_CHECK(ParseTuple, pcmcia_parse_tuple(handle, &tuple, &parse));
  
  link->conf.ConfigBase = parse.config.base;
  link->conf.Present    = parse.config.rmask[0];
  
  link->state |= DEV_CONFIG;

  CS_CHECK(GetConfigurationInfo, pcmcia_get_configuration_info(handle, &conf));
  
  tuple.DesiredTuple = CISTPL_CFTABLE_ENTRY;
  tuple.Attributes   = 0;
  CS_CHECK(GetFirstTuple, pcmcia_get_first_tuple(handle, &tuple));
  while (1) 
  {
    cistpl_cftable_entry_t *cfg = &parse.cftable_entry;
  
    if (pcmcia_get_tuple_data(handle, &tuple) != 0 || pcmcia_parse_tuple(handle, &tuple, &parse) != 0)
      goto next_entry;

    card = link->priv;

    if (cfg->io.nwin > 0) 
    {
      cistpl_io_t *io = &cfg->io;
      
      link->conf.Attributes  = CONF_ENABLE_IRQ;
      link->conf.IntType     = INT_MEMORY_AND_IO;
      link->conf.Vcc         = cfg->vcc.param[CISTPL_POWER_VNOM] / 10000;
      link->conf.Vpp1        = 0; // cfg->vpp1.param[CISTPL_POWER_VNOM] / 10000;
      link->conf.Vpp2        = 0; // cfg->vpp2.param[CISTPL_POWER_VNOM] / 10000;
      link->conf.ConfigIndex = cfg->index;
      
      link->irq.Attributes   = IRQ_TYPE_DYNAMIC_SHARING;
      link->irq.IRQInfo1     = IRQ_LEVEL_ID;
      link->irq.Handler      = NULL; // we use our own handler
      
      link->io.BasePort1     = io->win[0].base;
      link->io.NumPorts1     = io->win[0].len;
      link->io.Attributes1   = IO_DATA_PATH_WIDTH_8; // only this kind of access is yet supported
      link->io.IOAddrLines   = io->flags &  CISTPL_IO_LINES_MASK;
      
      if (pcmcia_request_io(link->handle, &link->io) == 0)
        break;
    }

    next_entry:
    CS_CHECK(GetNextTuple, pcmcia_get_next_tuple(handle, &tuple));
  }
  
  CS_CHECK(RequestIRQ, pcmcia_request_irq(handle, &link->irq));
  CS_CHECK(RequestConfiguration, pcmcia_request_configuration(handle, &link->conf));

  DPRINTK(KERN_DEBUG "%s: pccard found: base1=0x%04x, size=%d, irq=%d\n", DEVICE_NAME, link->io.BasePort1, link->io.NumPorts1, link->irq.AssignedIRQ);

  // init (cardmgr) devices associated with that card (is that necessary?)
  card->node.major = pcan_drv.nMajor;
  card->node.minor = PCCARD_MINOR_BASE;
  strcpy(card->node.dev_name, DEVICE_NAME);
  link->dev = &card->node;
  
  // create device descriptors associated with the card
  last_ret = pccard_create_all_devices(link);
  if (last_ret)
    goto fail;
  
  pccard_show_firmware_version(link->priv);  
  
  // enable power to connector
  pccard_enable_CAN_power(link->priv);
  
  link->state &= ~DEV_CONFIG_PENDING;
  return 0;
    
  fail:
  // remove card virtually
  pccard_plugout(link);
  
  cs_error(link->handle, last_fn, last_ret);
  
  link->state &= ~DEV_CONFIG_PENDING;
  
  DPRINTK(KERN_DEBUG "%s: pccard_plugin() failed!\n", DEVICE_NAME);
  return -1;
} // pccard_plugin()

//****************************************************************************
// detach entry
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
static void pccard_detach(dev_link_t *link)
{
  dev_link_t **plink;
  
  DPRINTK(KERN_DEBUG "%s: pccard_detach()\n", DEVICE_NAME);
  
  // test if the link is part of the linked list of pccard devices
  for (plink = &pccard_card_list; *plink; plink = &(*plink)->next)
    if (*plink == link)
      break;
  if (*plink == NULL)
    return;
    
  // do not finish stale config
  if (link->state & DEV_STALE_CONFIG) 
    return;
  
  // un-register
  if (link->handle)
  {
    int ret;
    
    ret = pcmcia_deregister_client(link->handle);
    if (ret != CS_SUCCESS)
      cs_error(link->handle, DeregisterClient, ret);
    else
      link->handle = NULL;
  }
  
  // unlink from list
  *plink = link->next;
  if (link->priv)
  {
    DPRINTK(KERN_DEBUG "%s: kfree(0x%p)\n", DEVICE_NAME, link->priv);
    kfree(link->priv);
    link->priv = NULL;
  }
}
#else 
static void pccard_detach(struct pcmcia_device *pcmcia_dev)
{
  dev_link_t *link = dev_to_instance(pcmcia_dev);
  
  DPRINTK(KERN_DEBUG "%s: pccard_detach()\n", DEVICE_NAME);
  
  // finish pending work
  if (link->priv)
  {
    pccard_plugout(link);
    
    DPRINTK(KERN_DEBUG "%s: kfree(0x%p)\n", DEVICE_NAME, link->priv);
    kfree(link->priv);
  }
}
#endif

//****************************************************************************
// attach entry
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
static int pccard_event(event_t event, int priority, event_callback_args_t *args);

static dev_link_t *pccard_attach(void)
#else
static int pccard_probe(struct pcmcia_device *pcmcia_dev)
#endif
{
  struct pcan_pccard *card;
  dev_link_t     *link = NULL;
  int ret = 0;
  #if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
  client_reg_t   client_reg;
  #endif
  
  DPRINTK(KERN_DEBUG "%s: pccard_probe or .._attach()\n", DEVICE_NAME);
  
  // allocate a device structure
  card = (struct pcan_pccard *)kmalloc(sizeof(struct pcan_pccard), GFP_KERNEL);
  DPRINTK(KERN_DEBUG "%s: kmalloc() = 0x%p\n", DEVICE_NAME, card);
  if (card == NULL)
  {
    DPRINTK(KERN_DEBUG "%s: can't allocate card memory\n", DEVICE_NAME);
    ret = -ENOMEM;
    goto fail;
  }
  
  memset(card, 0, sizeof(*card));  
  
  // init pcmcia card specific parts
  link = &card->link;
  link->priv  = card;
  link->state = 0;
  
  #if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
  // insert into linked list of pccard devices - current lonly element
  link->next        = pccard_card_list;
  pccard_card_list  = link;
  
  // register client to card services
  #if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,13)
  client_reg.EventMask = CS_EVENT_CARD_INSERTION | CS_EVENT_CARD_REMOVAL |
	                 CS_EVENT_RESET_PHYSICAL | CS_EVENT_CARD_RESET |
	                 CS_EVENT_PM_SUSPEND     | CS_EVENT_PM_RESUME;
  client_reg.event_handler = &pccard_event;
  #endif
  client_reg.dev_info   = &pccard_info;
  client_reg.Version    = 0x0210; // version 2.1
  client_reg.event_callback_args.client_data = link; 
  ret = pcmcia_register_client(&link->handle, &client_reg);
  if (ret != CS_SUCCESS)
  {
    cs_error(link->handle, RegisterClient, ret);
    pccard_detach(link);
    DPRINTK(KERN_DEBUG "%s: kfree(0x%p)\n", DEVICE_NAME, card);
    kfree(card);
    link = NULL;
    goto fail;
  }
  
  fail:
  if (link != NULL)
  {
    DPRINTK(KERN_DEBUG "%s: pccard_attach() is OK\n", DEVICE_NAME);
  }
  
  return link;
  #else
  link->handle = pcmcia_dev;
  pcmcia_dev->instance = link;
  
  pccard_plugin(link);
  
  fail:
  DPRINTK(KERN_DEBUG "%s: pccard_probe() = %d\n", DEVICE_NAME, ret);
  return ret;
  #endif
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
//****************************************************************************
// is called at each card state change
static int pccard_event(event_t event, int priority, event_callback_args_t *args)
{
  dev_link_t     *link = (dev_link_t *)args->client_data;

  DPRINTK(KERN_DEBUG "%s: pccard_event(%d)\n", DEVICE_NAME, event);
  
  switch (event)
  {
    case CS_EVENT_EJECTION_REQUEST:
    break;
    
    case CS_EVENT_CARD_INSERTION:
    pccard_plugin(link);
    break;
    
    case CS_EVENT_CARD_REMOVAL:
    pccard_plugout(link);
    break;
    
    case CS_EVENT_PM_SUSPEND:
    case CS_EVENT_PM_RESUME:
    break;
    
    case CS_EVENT_RESET_PHYSICAL:
    break;
    
    case CS_EVENT_CARD_RESET:
    break;
  
    default:
    break;
  } 
  
  return 0;
}
#else
static int pccard_suspend(struct pcmcia_device *pcmcia_dev)
{
  return 0;
}

static int pccard_resume(struct pcmcia_device *pcmcia_dev)
{
  return 0;
}
#endif

//****************************************************************************
// fill struct pcmcia_driver and register it
static int pccard_init(void)
{
  DPRINTK(KERN_DEBUG "%s: pccard_init()\n", DEVICE_NAME);

  memset (&pcan_drv.pccarddrv, 0, sizeof(pcan_drv.pccarddrv));
  
  pcan_drv.pccarddrv.owner      = THIS_MODULE;
  pcan_drv.pccarddrv.drv.name   = DEVICE_NAME;
  
  #if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,13)
  pcan_drv.pccarddrv.attach     = pccard_attach;
  pcan_drv.pccarddrv.detach     = pccard_detach;
  #endif
  
  #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,13) && LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
  pcan_drv.pccarddrv.id_table   = pccard_id_table;
  pcan_drv.pccarddrv.event      = pccard_event;
  pcan_drv.pccarddrv.attach     = pccard_attach;
  pcan_drv.pccarddrv.detach     = pccard_detach;
  #endif
  
  #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16)
  pcan_drv.pccarddrv.id_table   = pccard_id_table;
  pcan_drv.pccarddrv.probe      = pccard_probe;
  pcan_drv.pccarddrv.remove     = pccard_detach;
  pcan_drv.pccarddrv.suspend    = pccard_suspend;
  pcan_drv.pccarddrv.resume     = pccard_resume;  
  #endif
  
  return pcmcia_register_driver(&pcan_drv.pccarddrv);
}

//****************************************************************************
// small interface to rest of driver, only register and deinit
void pcan_pccard_deinit(void)
{
  DPRINTK(KERN_DEBUG "%s: pcan_pccard_deinit()\n", DEVICE_NAME);
  
  #if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
  {
    dev_link_t *link;
    
    // remove rest of linked (stale) devices 
    for (link = pccard_card_list; link; )
    {
      dev_link_t *tmp_link = link;
      
      link = link->next;
      
      pccard_plugout(tmp_link);
      
      if (tmp_link->handle)
        pcmcia_deregister_client(tmp_link->handle);
        
      // free and remove from list
      if (tmp_link->priv)
      {
        DPRINTK(KERN_DEBUG "%s: kfree(0x%p)\n", DEVICE_NAME, tmp_link->priv);
        kfree(tmp_link->priv);
      }
    }
  }
  #endif
  
  if (!strcmp(pcan_drv.pccarddrv.drv.name, DEVICE_NAME))
  {
    // then it was registered
    // unregister pccard parts, makes a plugout of registered devices
    pcmcia_unregister_driver(&pcan_drv.pccarddrv);
  }
}

int  pcan_pccard_register_devices(void)
{
  int ret;
  
  if (!(ret = pccard_init()))
  {
    DPRINTK(KERN_DEBUG "%s: pcan_pccard_register_devices() is OK\n", DEVICE_NAME);  
  }
  
  return ret; 
}

// fin

